/*
 * Implantation de la salle des ventes
 *
 * @author Pierre-Vincent ARNAUD, Lela TOURSI, Stphanie CADEAU, Anne VERNET, Audrey JACCARD
 *
 */

import java.lang.*;
import java.util.*;
import SalleEnchere.*;
import SalleEnchere.SalleVentePackage.*;
import VenteAuxEncheres.*;
import org.omg.CORBA.*;
import java.io.*;
import easyNotify.*;
import NommageFacile.*;

/**
 * Cette classe est une implantation de l'IDL de la salle des ventes
 */
public class SalleVenteImpl extends SalleVentePOA implements Timerizable, SalleVenteOperations {
    /**
     * Attribut indiquant le nom de la machine du service d'vnements
     */
    private String machineEv="quad";
    /**
     * Attribut indiquant le port du service d'vnements
     */
    private int portEv=10001;
    /**
     * Attribut indiquant le canal du service d'vnements
     */
    private int canalEv;
    /**
     * Attribut contenant le catalogue des articles
     */    
    private Catalogue catalogue;
    /**
     * Attribut indiquant l'tat de l'enchre
     */
    private Enchere etatEnchere;
    /**
     * Attribut contenant une table de hachage pour retrouver les informations des clients en fonction de leur identificateur
     */
    private Hashtable clients;
    /**
     * Attribut contenant un compteur pour les identificateurs des clients
     */
    private int dernierId;
    /**
     * Attribut indiquant la date de dbut de la vente
     */
    private Date dateDebut;
    /**
     * Attribut contenant le timer
     */
    private Timer timer;
    /**
     * Attribut indiquant si la vente a commenc
     */
    private boolean commence;
    /**
     * Attribut indiquant le dlai entre les actions de la salle (en millisecondes)
     */
    private int delai;
    /**
     * Attribut indiquant le produit courant
     */
    private Produit produitCourant;
    /**
     * Attribut indiquant le numro du produit courant
     */
    private int numProduit;
    /**
     * Attribut contenant un thread pour le timer
     */
    private Thread threadTimer;
    /**
     * Attribut contenant la table des prix minimum des articles
     */
    private int[] prixMin;
    /**
     * Attribut contenant le pas d'enchre pour les articles
     */
    private int[] pas;
    /**
     * Attribut indiquant le nom de la salle des ventes sur le service de nommage
     */
    private String nomObjet;
    /**
     * Attribut contenant un fournisseur d'vnements
     */
    EasyNotifyStructuredPushSupplier notify;
    /**
     * Attribut contenant l'ORB du serveur
     */
    ORB orb;
    
    /**
     * Constructeur de la salle des ventes
     * @param datedeb Date de dbut de la vente
     * @param nomFic Nom du fichier contenant le catalogue
     * @param orb ORB du serveur
     * @param nomObjet Nom du serveur auprs du service de nommage
     */
    public SalleVenteImpl(Date datedeb, String nomFic, ORB orb, String nomObjet) {
	clients=new Hashtable();
	dateDebut=datedeb;
	commence=false;
	delai=3000;
	lireFichierCatalogue(nomFic);
	etatEnchere=new Enchere();
	this.orb=orb;
	this.nomObjet=nomObjet;
	try {
	    /* Cration du fournisseur d'vnements */
	  //notify=new EasyNotifyStructuredPushSupplier(orb,machineEv,new Integer(portEv).toString(),canalEv);
	    notify=new EasyNotifyStructuredPushSupplier(orb,machineEv,new Integer(portEv).toString());
	    notify.connect();
	    /* Rcupration du canal affect par le service d'vnements */
	    canalEv=notify.getChannelId().value;
	    System.out.println("===> Canal choisi : "+canalEv+"\n");
	}
	catch(Exception e) {
	    System.out.println(e);
	    erreur("Impossible de crer un fournisseur d'vnements");
	    System.exit(1);
	}
	/* Cration d'un timer pour calculer le temps restant avant le dbut de la vente */
	timer=new Timer(new Date(),dateDebut.getTime()-new Date().getTime(),this);
	/* Lancement du thread du timer */
	threadTimer=new Thread(timer);
	threadTimer.start();
    }

    /**
     * Cette mthode permet de renvoyer un message d'erreur
     * @param err Chane de caractre expliquant l'erreur
     */
    private void erreur(String err) {
	System.err.println("===> !!! "+err+"\n");
    }

    /**
     * Cette mthode permet d'initialiser le catalogue  partir d'un fichier
     * @param nomFic Nom du fichier  lire
     */
    public void lireFichierCatalogue(String nomFic) {
	File fic;
	FileReader fluxFic;
	BufferedReader fluxTamp;
	int i;
	String nom=new String();
	String description=new String();
	int prix=0;
	int taille;
	Produit produits[];
	try {
	    fic=new File(nomFic);
	    fluxFic=new FileReader(fic);
	    fluxTamp=new BufferedReader(fluxFic);	   
	    String ch;
	    taille=0;
	    /* Boucle rcuprant le nombre d'articles */
	    while((ch=fluxTamp.readLine())!=null) {
		StringTokenizer st=new StringTokenizer(ch," ");
		if(st.hasMoreTokens()) {
		    taille++;
		}
	    }
	    produits=new Produit[taille];
	    prixMin=new int[taille];
	    pas=new int[taille];
	    i=0;
	    fluxFic=new FileReader(fic);
	    fluxTamp=new BufferedReader(fluxFic);
	    while((ch=fluxTamp.readLine())!=null) {
		StringTokenizer st=new StringTokenizer(ch," ");
		if(st.hasMoreTokens()) {
		    /* Rcupration du nom */
		    nom=new String(st.nextToken());
		    if (st.hasMoreTokens()) {
			/* Rcupration de la description */
			description=new String(st.nextToken());
			if (st.hasMoreTokens()) {
			    /* Rcupration du prix */
			    prix=new Integer(st.nextToken()).intValue();
			    if (st.hasMoreTokens()) {
				/* Rcupration du prix minimal */
				prixMin[i]=new Integer(st.nextToken()).intValue();
				if (st.hasMoreTokens()) {
				    /* Rcupration du pas */
				    pas[i]=new Integer(st.nextToken()).intValue();
				}
				else {
				    erreur("Pas de baisse attendu");
				}
			    } 
			    else {
				erreur("Prix minimal attendu");
			    }
			}
			else {
			    erreur("Prix attendu");
			}
		    }
		    else {
			erreur("Descriptif attendu");
		    }
		}
		Produit p=new Produit(i,nom,description,prix,false,0);
		produits[i]=p;
		i++;
	    }
	    catalogue=new Catalogue((int)tempsRestant(),"notre salle de vente",produits);
	}
	catch(Exception ex) {
	    erreur(ex.getClass().getName());
	    erreur("Pas de fichier "+nomFic);
	    System.exit(1);
	}
    }

    /**
     * Cette mthode permet de passer  un nouveau produit
     */
    private void nouveauProduit() {
	/* remise  zro du timer */
	timer.zero();
	if(numProduit>=catalogue.produits.length) {
	    /* Si tous les produits sont passs */
	    try {
		/* envoi d'un vnement pour la fermeture de la salle */
		notify.push(new SdvEvent(orb,etatEnchere,"","FERMETURE"));
	    }
	    catch(Exception e) {
		erreur("Impossible d'envoyer l'vnement FERMETURE");
		System.exit(1);
	    }
	    try {
		/* dconnexion du service d'vnements */
		notify.disconnect();
	    }
	    catch(Exception e) {
		erreur("Impossible de se dconnecter du service d'vnements");
		System.exit(1);
	    }
	    /* dsinscription du service de nommage */
	    EasyNaming.desinscriptionObjet(orb,nomObjet);
	    System.out.println("===> Fermeture de la salle\n");
	    System.exit(0);
	}
	else {
	    /* Sinon */
	    /* On passe  un nouveau produit */
	    produitCourant=catalogue.produits[numProduit];
	    etatEnchere.leStatut=Statut.enBaisse;
	    etatEnchere.prixCourant=produitCourant.prixDepart;
	    etatEnchere.idProduitCourant=produitCourant.id;
	    try {
		/* envoi d'un vnement pour la mise en vente d'un nouveau produit */
		notify.push(new SdvEvent(orb,etatEnchere,"","NOUVEAU PRODUIT"));
	    }
	    catch(Exception e) {
		erreur("Impossible d'envoyer l'vnement NOUVEAU PRODUIT");
		System.exit(1);
	    }
	    System.out.println("===> Nouveau produit mis en vente: "+produitCourant.description+"  "+produitCourant.prixDepart+"\n");
	}
    }

    /**
     * Cette mthode est lance  chaque top du timer
     */
    public void timer() {
	if (!commence) {
	    /* Si la vente n'a pas commenc */
	    /* On commence la vente (puisqu'il s'agit du premier top du timer, indiquant que la date de dbut est arrive) */
	    System.out.println("===> Dbut de la vente\n");
	    try {
	      /* Ici il faut que l'on renseigne les champs de l'tat de l'enchre pour qu'elle puisse tre traite par l'vnement (qui s'intresse  ces champs) */
	      etatEnchere=new Enchere(-1,-1,-1,Statut.enBaisse);
	      notify.push(new SdvEvent(orb,etatEnchere,"","DEBUT"));
	    }
	    catch(Exception e) {
		erreur("Impossible d'envoyer l'vnement DEBUT");
		System.exit(1);
	    }
	    commence=true;
	    /* on modifie le dlai du timer pour qu'il corresponde au dlai entre des actions */
	    timer.changerDelai(delai);
	    timer.zero();
	    numProduit=0;
	    nouveauProduit();
	}
	else {
	    /* Sinon */
	    if(etatEnchere.leStatut==Statut.enBaisse) {
		/* Si le statut est "en baisse" */
		if(etatEnchere.prixCourant-pas[numProduit]<=prixMin[numProduit]) {
		    /* Si le prix atteint est en dessous du seuil */
		    try {
			notify.push(new SdvEvent(orb,etatEnchere,"","PRODUIT RETIRE"));
		    }
		    catch(Exception e) {
			erreur("Impossible d'envoyer l'vnement PRODUIT RETIRE");
			System.exit(1);
		    }
		    /* On retire le produit */
		    System.out.println("===> "+produitCourant.description+" retir\n");
		    numProduit++;
		    nouveauProduit();
		}
		else {
		    /* Sinon */
		    etatEnchere.prixCourant=etatEnchere.prixCourant-pas[numProduit];
		    /* On dcrmente le prix */
		    try {
			notify.push(new SdvEvent(orb,etatEnchere,"","PRODUIT EN BAISSE"));
		    }
		    catch(Exception e) {
			erreur("Impossible d'envoyer l'vnement PRODUIT EN BAISSE");
			System.exit(1);
		    }
		    System.out.println("===> "+produitCourant.description+" en baisse  "+etatEnchere.prixCourant+"\n");
		}
	    }
	    else if(etatEnchere.leStatut==Statut.encheri) {
		/* Sinon si le statut est "enchri" */
		/* On passe  "une fois" */
		etatEnchere.leStatut=Statut.uneFois;
		try {
		    notify.push(new SdvEvent(orb,etatEnchere,"","UNE FOIS"));
		}
		catch(Exception e) {
		    erreur("Impossible d'envoyer l'vnement UNE FOIS");
		    System.exit(1);
		}
		System.out.println("===> "+produitCourant.description+": une fois!"+"\n");
	    }
	    else if(etatEnchere.leStatut==Statut.uneFois) {
		/* Sinon si le statut est "une fois" */
		/* On passe  "deux fois" */
		etatEnchere.leStatut=Statut.deuxFois;
		try {
		    notify.push(new SdvEvent(orb,etatEnchere,"","DEUX FOIS"));
		}
		catch(Exception e) {
		    erreur("Impossible d'envoyer l'vnement DEUX FOIS");
		    System.exit(1);
		}
		System.out.println("===> "+produitCourant.description+": deux fois!\n");
	    }
	    else if(etatEnchere.leStatut==Statut.deuxFois) {
		/* Sinon si le statut est  "deux fois" */
		/* On passe  "trois fois" */
		etatEnchere.leStatut=Statut.troisFois;
		try {
		    notify.push(new SdvEvent(orb,etatEnchere,"","TROIS FOIS"));
		}
		catch(Exception e) {
		    erreur("Impossible d'envoyer l'vnement TROIS FOIS");
		    System.exit(1);
		}
		System.out.println("===> "+produitCourant.description+": trois fois!\n");
	    }
	    else if(etatEnchere.leStatut==Statut.troisFois) {
		/* Sinon si le statut est  "trois fois" */
		/* On passe  "adjug vendu" */
		etatEnchere.leStatut=Statut.adjuge;
		produitCourant.prixVente=etatEnchere.prixCourant;
		/* Le produit est vendu */
		produitCourant.vendu=true;
		try {
		    notify.push(new SdvEvent(orb,etatEnchere,"","ADJUGE VENDU"));
		}
		catch(Exception e) {
		    erreur("Impossible d'envoyer l'vnement ADJUGE VENDU");
		    System.exit(1);
		}
		System.out.println("===> "+produitCourant.description+": adjug vendu!\n");
		numProduit++;
		nouveauProduit();
	    }
	}
    }

    /**
     * Cette mthode permet de tester si l'on est arriv au dernier produit
     * @return Un boolen contenant la valeur du test
     */
    private boolean dernierProduit() {
	return catalogue.produits[catalogue.produits.length-1].id==etatEnchere.idProduitCourant;
    }

    /**
     * Cette mthode permet de renvoyer le temps restant avant le dbut de la vente
     * @return Un entier long contenant ce temps
     */
    private long tempsRestant() {
	return new Date().getTime()-dateDebut.getTime();
    }

    /**
     * Cette mthode permet de renvoyer le catalogue des articles
     * @return Le catalogue de la salle de vente
     */
    public Catalogue getCatalogue() {
	System.out.println("===> Un client demande le catalogue\n");
	catalogue.tempsRestant=(int)tempsRestant();
	return catalogue;
    }
    
    /**
     * Cette mthode permet l'entre d'un client dans la salle de vente
     * @param p La personne dsirant entrer
     * @param idPersonne L'identificateur de la personne (mis  jour par la mthode)
     * @param idEV L'identificateur du service d'vnement (mis  jour par la mthode)
     * @return L'enchre courante
     * @throws SalleFermee Une exception indiquant que la salle est ferme
     */
  public Enchere entree(Personne p, IntHolder idPersonne, EventServiceIdHolder idEV) throws SalleFermee {
    System.out.println("===> Un client essaie d'entrer ("+p.prenom+" "+p.nom+")\n");
    if(etatEnchere.leStatut==Statut.adjuge&&dernierProduit()) {
	    System.out.println("===> "+p.prenom+" "+p.nom+" est refus (salle ferme)\n");
	    throw new SalleFermee();
	}
	else {
	    dernierId++;
	    clients.put(new Integer(dernierId),p);
	    idPersonne.value=dernierId;
	    idEV.value=new EventServiceId(canalEv,portEv,machineEv);
	    return etatEnchere;
	}
    }

    /**
     * Cette mthode permet  un client d'enchrir
     * @param idPersonne L'identificateur de la personne
     * @param idProduit L'identificateur du produit
     * @param prix Le montant de l'enchre
     * @throws EnchereBasse Une exception indiquant que l'enchre est trop basse
     * @throws MauvaisProduit Une exception indiquant que l'enchre porte sur un mauvais produit
     */
    public void encherir(int idPersonne, int idProduit, int prix) throws EnchereBasse, MauvaisProduit {
	if(etatEnchere.leStatut==Statut.adjuge&&dernierProduit()) {
	    /* Si les enchres sont finies */
	    /* En fait il devrait y avoir une exception indiquant que la salle est ferme */
	}
	else if(etatEnchere.prixCourant>=prix) {
	    /* Sinon si l'enchre est trop basse */
	    System.out.println("===> Enchre trop basse de "+idPersonne+"\n");
	    throw new EnchereBasse();
	}
	else if(etatEnchere.idProduitCourant!=idProduit) {
	    /* Sinon s'il s'agit du mauvais produit */
	    System.out.println("===> Enchre sur un mauvais produit de "+idPersonne+"\n");
	    throw new MauvaisProduit();
	}
	else {
	    /* Sinon, l'enchre est acceptable */
	    etatEnchere.leStatut=Statut.encheri;
	    /* On met  jour l'tat de l'enchre */
	    etatEnchere=new Enchere(idPersonne,etatEnchere.idProduitCourant,prix,etatEnchere.leStatut);
	    try {
		notify.push(new SdvEvent(orb,etatEnchere,"","ENCHERE"));
	    }
	    catch(Exception e) {
		erreur("Impossible d'envoyer l'vnement ENCHERE");
		System.exit(1);
	    }
	    System.out.println("===> Enchre accepte de "+idPersonne+" ("+prix+")\n");
	    /* Mise  zro du timer */
	    timer.zero();
	}
    }

    /**
     * Cette mthode permet  un client de sortir
     * @param idPersonne L'identificateur du client
     */
    public void sortie(int idPersonne) {
	/* On supprime l'identificateur du client de la table de hachage */
	clients.remove(new Integer(idPersonne));
	System.out.println("===> Sortie de "+idPersonne+"\n");
    }
}
